/* sieve_sump.cs is to sum the primes to determine how many */
/* primes there are in 2^30+1 numbers.  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h> 
#include <cscomm.h> 
#include <cm/cmfs.h>
#include <cm/cm_file.h>

#define CSIZE 32768                        /* column size */
#define RSIZE 16384                        /* row size */

shape [CSIZE]sieve;    /* Shape for sieve representation */
bool:sieve is_prime;    /* True if number is prime */
bool:sieve mk[RSIZE];   /* bit vector with primes marked as true */
long:sieve num=0;
int:sieve number=0; 
double:sieve x_log_x;
int:sieve result;

void num_primes() 
 {  int r; 
    long total_num;
   
    total_num = 0; 
 
    with(sieve) { 
     for (r = 0; r < RSIZE; r++)
      num += mk[r];  
     
      total_num = (+= num);
     
      printf("\n");   
      printf("total number of odd primes in 2^30+1 is %d\n",total_num);
      printf("\n");  
   } /* end with */ 
}

void scan_sum()
 { int r,t,n;  
 
   with(sieve)
     {
        result = 0; 
        for (r=0; r<RSIZE; r++)
          number += mk[r];
        
        result = scan(number,0,CMC_combiner_add,CMC_upward,
          CMC_none,CMC_no_field,CMC_inclusive);  
        
        printf("\n\n");  
        printf("**** Scan results on data ****\n");     

        for(t=0; t<CSIZE; t=t+163)
         { n = (CSIZE*(t+1))+1;         
           printf("%d\n",n);
           printf("%d\n",[t]result);
          }  /* end for */ 
       
           n = (CSIZE*CSIZE)+1 ;  
           printf("%d\n",n);
           printf("%d\n",[32767]result); 
    } /* end with */
}

void ln_function()
 {  double x,ln_result; 
    int r,n;
  
    with(sieve)
     {  
  
        printf("\n\n"); 
        printf("**** Results of x/ln(x) ****\n");  
    
        for(r=0; r<CSIZE; r++)
         { x = (CSIZE*(r+1))+1;
           ln_result = log(x); 
           [r]x_log_x = x/ln_result;
         } 
   
        for(r=0; r<CSIZE; r=r+163)
         {  n = (CSIZE*(r+1))+1;
            printf("%d\n",n);
            printf("%f\n",[r]x_log_x);
         }
         
        n = (CSIZE*CSIZE)+1;
        printf("%d\n%",n); 
        printf("%f\n",[32767]x_log_x);
    } /* end with */   
}  

void ratio_ln_result()
 { int r,n;
   float:sieve ratio;

   with(sieve)
    { 
      ratio = x_log_x / result;  
      
      printf("\n\n");
      printf("**** Ratio of x/ln(x) / results ****\n");
     
      for (r=0; r < CSIZE; r=r+163) 
       {  n = (CSIZE*(r+1))+1;  
          printf("%d\n",n);
          printf("%f\n",[r]ratio);
       }  /* end for */

      n = (CSIZE*CSIZE)+1;
      printf("%d\n",n); 
      printf("%f\n",[32767]ratio);
    } /* end with */
} 

void differ_result_ln()
 { int r,n;
   double:sieve differ;

   with(sieve)
    { 
      differ = result - x_log_x;
 
      printf("\n\n");
      printf("**** Difference of result - x/ln(x) ****\n");

      for (r=0; r < CSIZE; r=r+163)
       {  n = (CSIZE*(r+1)) + 1; 
          printf("%d\n",n);
          printf("%f\n",[r]differ);
       }  /* end for */

      n = (CSIZE * CSIZE)+1 ;
      printf("%d\n%f\n",n,[32767]differ);
    } /* end with */
}

void checkpoint()
 {
   int mark,ispr,dvfd;
   char fnm[50], type[2];

   ispr = boolsizeof(bool:sieve);
   mark = RSIZE * ispr;

   printf("r,w> ");  scanf("%s",type);
   printf("Checkpoint file> ");  scanf("%s", fnm);

   with (sieve)
   { if ( ( type[0] == 'w' ) && ( type[1] == 0 ) )
      { /* Checkpoint in user-specified files */
        CM_start_timer(1);
        dvfd = CMFS_creat(fnm, 0600);
        CMFS_write_file(dvfd, &mk[0], mark);
        CMFS_write_file(dvfd, &is_prime, ispr);
        CMFS_close(dvfd);
        CM_stop_timer(1);  CM_reset_timer();
      }
    else if ( ( type[0] == 'r' ) && ( type[1] == 0 ) )
      { /* Resume from user-specified files - CM must have been */
        /* initialized with proper geometry by previous run.    */
        CM_start_timer(1);
        dvfd = CMFS_open(fnm, CMFS_O_RDONLY);
        CMFS_read_file(dvfd, &mk[0], mark);
        CMFS_read_file(dvfd, &is_prime,  ispr);
        CMFS_close(dvfd);
        CM_stop_timer(1);  CM_reset_timer();
      }
   }
}

/* Main program.  Commands are:
   x  - checkpointing
   s  - sum determine number of primes 
   c  - scan_sum
   l  - ln_function
   r  - ratio_result_ln
   d  - differ_result_ln 
   q  - quit           */

void main()
 { char c[2];
   
   while ( printf("sieve>"), scanf("%1s", c), c[0] != 'q' )
     switch (c[0])
      { case 'x':
          checkpoint();
          break;
        case 's': 
          CM_start_timer(1);
          num_primes();
          CM_stop_timer(1);  CM_reset_timer();
          break;
        case 'c':
          CM_start_timer(1);
          scan_sum();
          CM_stop_timer(1);  CM_reset_timer();
          break;
        case 'l':
          CM_start_timer(1);
          ln_function();
          CM_stop_timer(1);  CM_reset_timer();
          break;
        case 'r':
          CM_start_timer(1);
          ratio_ln_result();
          CM_stop_timer(1);  CM_reset_timer();
          break;
        case 'd':
          CM_start_timer(1);
          differ_result_ln();
          CM_stop_timer(1);  CM_reset_timer();  
          break;
        default:   printf("Bad input - ignored\n"); break;
     }
} 
